/*---------------------------------------------------------------*/
/* Mechanical Stroboscope Controller             (C) ChaN, 2015  */
/*---------------------------------------------------------------*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>

FUSES = {0xEF, 0xC9};		/* ATmega32 fuses: Low, High */


#define K_POWER	1
#define K_UP	2
#define K_DOWN	4

#define	F_CTR		(F_CPU / 8)	/* Timer count rate */
#define N_PPS		8			/* Number of pulses per shutter */
#define N_KV		5			/* Motor EMF constant (x10mV/shutters/sec) */
#define GAIN_P		500
#define GAIN_I		20
#define GAIN_D		0
#define SET_PWM(v)	OCR0 = 255 - (v)


#define BATT_OK	3650		/* Minimum startup voltage */
#define BATT_LO	3500		/* Shutdown voltage */


volatile uint8_t Disp7seg[4], DispOn;
volatile uint8_t Cmd, Btn, Tmr100;
volatile uint16_t Vin;

volatile uint8_t ServoOn;	/* 0:Disable servo, 1:Eneble servo */
volatile int16_t ServoVal;	/* Commanded rotation speed (x0.1 shutters/sec) */
volatile int8_t SpsMul;		/* Number of pulses captured */
volatile int32_t SpsCtr;	/* Sum of the captured time */

uint16_t EE_Value EEMEM;



/*-------------------------------------------------------*/
/* System wakeup process                                 */
/*-------------------------------------------------------*/

ISR(INT0_vect)
{
	GICR = 0;	/* Disable INT0 */
}



/*-------------------------------------------------------*/
/* Timing pulse process                                  */
/*-------------------------------------------------------*/

ISR(TIMER1_CAPT_vect)
{
	static uint16_t pv;
	uint16_t icr;


	icr = ICR1;
	SpsCtr += icr - pv; pv = icr;
	SpsMul++;
}



/*-------------------------------------------------------*/
/* 500Hz ivt process                                     */
/*-------------------------------------------------------*/

ISR(TIMER2_COMP_vect)
{
	/* Display drive and Key scan */
	{
		static uint8_t lc, tc, lbt;
		uint16_t vin;
		uint8_t n, v, k, kp;

		PORTA &= 0x0F;
		if (DispOn) {
			lc = (lc + 1) % sizeof Disp7seg;
			PORTC = Disp7seg[lc];
			PORTA |= 0x10 << lc;
		}

		if (++tc >= 5) {
			tc = 0;

			n = Tmr100;
			if (n) Tmr100 = n - 1;

			k = Btn; Btn = kp = ~PIND >> 2 & 7;
			v = (k ^ kp) & kp;
			if (v) Cmd = v;

			Vin = vin = 1024UL * 1220 / ADC;
			ADMUX = _BV(REFS0)|0b11110;
			ADCSRA = _BV(ADEN)|_BV(ADSC)|_BV(ADIF)|0b111;
			if (vin < BATT_LO) {
				if (lbt >= 100)
					Cmd = K_POWER;
				else
					lbt++;
			} else {
				lbt = 0;
			}
		}
	}

	/* Servo control */
	{
		static int32_t ireg, spd;
		static uint8_t div, pwmi, pwmf;
		int8_t rp, clip;
		uint16_t b;
		int32_t err, perr, trq, vout;
		uint32_t ctr;

		if (++div >= 10) {	/* 50 times per second for servo process */
			div = 0;

			/* Get rotation velocity (spd: *0.01sps) */
			cli();
			rp = SpsMul; SpsMul = 0;
			ctr = SpsCtr; SpsCtr = 0;
			sei();
			if (rp)	/* Get the velocity with the counter data */
				spd = 100L * F_CTR / N_PPS * rp / ctr;

			/* Servo process */
			if (ServoOn) {
				err = ServoVal - spd;			/* Get velocity error */
				trq = err * GAIN_P / 1024 + ireg * GAIN_I / 1024 + (err - perr) * GAIN_D / 1024;	/* Get torque command */
				perr = err;
				vout = trq + spd * N_KV / 100;	/* Get output voltage with Back-EMF compensation */

				/* Set output voltage */
				clip = 0;
				if (vout > 1030) { vout = 1030; clip = 1; }
				if (vout < 0) { vout = 0; clip = 2; }
				b = 1000UL * 256 / 1100 * vout / 100;
				pwmi = b / 10; pwmf = b % 10;

				/* Error integration with anti-windup */
				if ((err > 0 && clip != 1) || (err < 0 && clip != 2))
					ireg += err;
			} else {
				ireg = 0;			/* Reset integrator */
			}
		}
		if (ServoOn) SET_PWM(div >= pwmf ? pwmi : pwmi + 1);	/* Set PWM out with 10-step dithering */
	}
}



void delay_ms (uint16_t ms)
{
	Tmr100 = ms / 10;
	while (Tmr100) ;
}



/*-------------------------------------------------------*/
/* Motor driver controls                                 */
/*-------------------------------------------------------*/


void motor_off (void)
{
	TCCR0 = 0;			/* Stop PWM output */
	PORTB &= ~_BV(4);	/* Disable DC-DC converter and photo sensor */
}




void motor_on (void)
{
	PORTB |= _BV(4);	/* Enable DC-DC converter and photo sensor */
	delay_ms(100);		/* Wait for power stable */

	/* Start PWM output */
	SET_PWM(0);
	TCCR0 = _BV(COM01)|_BV(WGM00)|_BV(WGM01)|0b001;
}




/*-------------------------------------------------------*/
/* Display controls                                      */
/*-------------------------------------------------------*/


static
void disp (int16_t val, int dp)
{
	static const uint8_t seg7[16] = { 0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07, 0x7F, 0x6F, 0x40, 0x73, 0x79, 0x76, 0x38, 0x00 };	/* 0123456789-PEHL  */
	uint8_t i;


	i = 4;
	do {
		Disp7seg[--i] = seg7[val % 10]; val /= 10;
	} while (i && val);
	if (i == 3) Disp7seg[--i] = seg7[0];
	if (dp) Disp7seg[2] ^= 0x80;
	while (i) Disp7seg[--i] = seg7[15];
}




static
void ioinit (void)
{
	/* Initialize GPIO */
	PORTA = 0x0F; DDRA = 0xF0;	/* Port A: L L L L u u u u */
	PORTB = 0xEE; DDRB = 0x19;	/* Port B: u u u L H u u L */
	PORTC = 0x00; DDRC = 0xFF;	/* Port C: L L L L L L L L */
	PORTD = 0xBF; DDRD = 0x00;	/* Port D: u z u u u u u u */

	/* Initialize Timer2 as 500Hz IVT */
	ASSR = 0;
	OCR2 = F_CPU / 128 / 500 - 1;
	TCCR2 = _BV(WGM21)|0b101;

	/* Initialize Timer1 as input capture */
	TCCR1A = 0;
	TCCR1B = _BV(ICNC1)|0b010;

	/* Eneble interrupts */
	TIMSK = _BV(OCIE2)|_BV(TICIE1);
	sei();
}



static
void enter_off_state (void)
{
	uint16_t vin;


	do {
		DispOn = 0;
		do {	/* Wait for center button released */
			delay_ms(10);
		} while (Btn & K_POWER);

		/* Enter MCU sleep mode */
		do {
			set_sleep_mode(SLEEP_MODE_PWR_DOWN);
			cli();
			ADCSRA = 0;				/* Shutdown ADC */
			GICR = _BV(INT0);		/* Enable INT0 in level triggered mode */
			sleep_enable();
			sei();
			sleep_cpu();			/* Enter power down mode */
			sleep_disable();
			delay_ms(20);			/* Start 20ms after wakeup */
		} while (~Btn & K_POWER);

		DispOn = 1;
		cli(); vin = Vin; sei();
		disp(vin / 10, 0);
		do {	/* Wait for center button released */
			delay_ms(10);
		} while (Btn & K_POWER);

	} while (vin < BATT_OK);

	Cmd = 0;
}




static
void set_val (int16_t spd)
{
	disp(spd, 1);
	spd *= 10;
	cli(); ServoVal = spd; sei();
}




static
void up_down (int16_t* val, uint8_t cmd)
{
	uint8_t dly = 30;
	uint16_t t = 0;


	do {
		if (cmd == K_UP) {
			if (*val < 1800) (*val)++;
		} else {
			if (*val > 50) (*val)--;
		}
		set_val(*val);

		if (t >= 70 && dly > 6) dly = 6;
		if (t >= 200 && dly > 1) dly = 1;
		if (dly != 1) t += dly;
		for (Tmr100 = dly; Tmr100; ) ;
	} while (!Cmd && (Btn & (K_UP | K_DOWN)));
}




/*-------------------------------------------------------*/
/* Main process                                          */
/*-------------------------------------------------------*/

int main (void)
{
	uint8_t cmd;
	int16_t v;


	ioinit();

	for (;;) {
		enter_off_state();

		v = eeprom_read_word(&EE_Value);
		if (v > 1800 || v < 50) v = 500;
		set_val(v);

		motor_on();
		SET_PWM(80);
		delay_ms(100);
		ServoOn = 1;

		for (;;) {
			while (!Cmd) ;
			cmd = Cmd; Cmd = 0;
			if (cmd == K_POWER) break;
			if (cmd == K_UP || cmd == K_DOWN) {
				up_down(&v, cmd);
			}
		}

		ServoOn = 0;
		motor_off();
		eeprom_write_word(&EE_Value, v);
	}
}

